extern "C" {
    #include <stdio.h>
    #include <math.h>
    #include <string.h>
    #include <stdlib.h>
    #include <assert.h>
    #include <time.h>
    #include "src/definitions.h"
    #include "src/functions.h"
    #include "src/routines.h"
}
